package DB;

import NameServer.*;
import RPC.*;

// Example of starting the DB server program.  Your version may be somewhat
// different depending on how you implement RPC.   See Test.TestRPC to see
// a similar class with comments.

public class DBServer {

    static final String outputHdr = "##  DB.DBServer: ";

    public static void
    main (String[] argv)
    {
	String       usageString = "usage: DBServer <NS-NAME> <LOCAL-NAME> <DB-NAME> <EPOCH>";
	String       nsName;
	String       localName;
	String       dbName;
	int          epoch;
	NameContextI rootContext, dbContext;
	
	try {
	    nsName = argv[0];
	    localName = argv[1];
	    dbName   = argv[2];
	    epoch  = Integer.parseInt (argv[3]);
	} catch (ArrayIndexOutOfBoundsException abe) {
	    System.out.println (usageString);
	    return;
	} catch (NumberFormatException nfe) {
	    System.out.println (usageString);
	    return;
	}
	
	try {

	    // start server thread and bind NameServerCP (using file)
	    System.out.println(outputHdr + "starting the local server.");
	    RPC.Server.start       (epoch);
	    RPC.Client aClient = new RPC.Client();
	    NameServerCP.bind("NameServer.NameServer", nsName);

	    // get root context from name server
	    System.out.println(outputHdr + "getting the root name context");
	    rootContext = NameServerCP.getRootContext ();

	    // binding name context CP
	    System.out.println(outputHdr + "binding name context client proxy");
	    NameContextCP.bind(nsName);

	    // create local name context
	    System.out.println(outputHdr + "creating local name context");
	    Server.LocalNameContext = NameContextCP.rootNameContext.addContext(localName);
	    dbContext = Server.LocalNameContext;
	    
	    // look for context named "DBServer"; create it if not found
	    //dbContext   = rootContext.lookup ("DBServer");
	    //if (dbContext == null) 
	    //	dbContext = rootContext.addContext ("DBServer");

	    // register the database classes with the dbContext
	    String hashval;
	    System.out.println(outputHdr + "registering classes");

	    hashval = String.valueOf(RPC.ClassID.Cell) + "." + String.valueOf(CellRPC.CLASS_VERSION);
	    RPC.Server.classRegistry.put(hashval, "DB.Cell");
	    RPC.Server.LocalNameContext.add("DB.Cell", RPC.Server.server.getComHandle());
	    System.out.println("\tregistered DB.Cell");

	    hashval = String.valueOf(RPC.ClassID.Database) + "." + String.valueOf(DatabaseRPC.CLASS_VERSION);
	    RPC.Server.classRegistry.put(hashval, "DB.Database");
	    RPC.Server.LocalNameContext.add("DB.Database", RPC.Server.server.getComHandle());
	    System.out.println("\tregistered DB.Database");

	    hashval = String.valueOf(RPC.ClassID.Index) + "." + String.valueOf(IndexRPC.CLASS_VERSION);
	    RPC.Server.classRegistry.put(hashval, "DB.Index");
	    RPC.Server.LocalNameContext.add("DB.Index", RPC.Server.server.getComHandle());
	    System.out.println("\tregistered DB.Index");

	    hashval = String.valueOf(RPC.ClassID.Node) + "." + String.valueOf(NodeRPC.CLASS_VERSION);
	    RPC.Server.classRegistry.put(hashval, "DB.Node");
	    RPC.Server.LocalNameContext.add("DB.Node", RPC.Server.server.getComHandle());
	    System.out.println("\tregistered DB.Node");

	    hashval = String.valueOf(RPC.ClassID.Table) + "." + String.valueOf(TableRPC.CLASS_VERSION);
	    RPC.Server.classRegistry.put(hashval, "DB.Table");
	    RPC.Server.LocalNameContext.add("DB.Table", RPC.Server.server.getComHandle());
	    System.out.println("\tregistered DB.Table");
	    
	    // add a database with provided name to the DBServer context
	    System.out.println(outputHdr + "adding a database");
	    dbContext.add (dbName, new Database(dbName));

	    // block the mainline thread
	    System.out.println(outputHdr + "\n\nwaiting for requests...\n\n\n");
	    RPC.Server.join ();
	    
	} catch (Exception ns) {
	    System.out.println (outputHdr + "Name server not found");
	    return;
	}
    }
}
