package DB.Test;

import DB.*;

// This class is an example of a test that adds some cells to the table,
// builds two indexes, and then prints the table out in the orders defined
// by these indexes.  Note that cell objects are created in two ways, using
// "new Cell()" and using "table.addNewCell()".  Be sure you understand the
// difference between these two calls and what is really going on here.
//
// To fully test your system, of course, you may want to add other test cases,
// this is just an example.  But, be sure that you can at least do this
// example and include its execution in what you hand in.
//

public class Test {
    static Index nameIndex;
    static Index valueIndex;

    static void
    buildIndexes (TableI table)
    {
	CellI cell;

	nameIndex  = new Index ();
	valueIndex = new Index ();
	for ( cell = table.getFirstCell(); cell!=null; cell = cell.getNext () ) {
	    nameIndex.insert  (cell.getName(),                    cell);
	    valueIndex.insert (Integer.toString(cell.getValue()), cell);
	}
    }

    static void
    createTable (DatabaseI db, String tableName)
    {
        TableI table;
        
        table = db.addNewTable (tableName);
        
	//
	// Put some data in table
	//
        table.addNewCell ("Mike", 80);
        table.addNewCell ("Linda", 73);
        table.addCell    (new Cell("Peter", 33));
	table.addNewCell ("Sharon", 76);
        table.addCell    (new Cell("Teagan", 97));
	table.addNewCell ("Erin", 99);

	//
	// Build name and value indexes
	//
	buildIndexes (table);

	//
	// Print
	//
	System.out.println      ("After Create");
	System.out.println      ("Table in Name Order");
	nameIndex.printInOrder  ();
	System.out.println      ("");
	System.out.println      ("Table in Value Order");
	valueIndex.printInOrder ();
        System.out.println      ("");
    }

    static void
    readTable (DatabaseI db, String tableName)
    {
	TableI table;

	table = db.lookupTable (tableName);
	if (table == null) {
	    System.out.println ("Table not found");
	} else {
	    buildIndexes (table);
	    System.out.println ("Read from Database");
	    System.out.println ("Table in Name Order");
	    nameIndex.printInOrder  ();
            System.out.println      ("");		
	    System.out.println ("Table in Value Order");
	    valueIndex.printInOrder ();
            System.out.println      ("");
	}
    }

    static void
    emptyTable (DatabaseI db, String tableName)
    {
	TableI table;

	table = db.lookupTable (tableName);

	if (table == null) {

	    System.out.println ("Table not found"); 

	} else {

	    //
	    // Delete entries from table by name
	    //
	    table.removeCell ( table.lookupCell ("Mike") );
	    table.removeCell ( table.lookupCell ("Linda") );
	    table.removeCell ( table.lookupCell ("Peter") );
	    table.removeCell ( table.lookupCell ("Sharon") );
            table.removeCell ( table.lookupCell ("Teagan") );
            table.removeCell ( table.lookupCell ("Erin") );

	    //
	    // Now build another index and print again to be sure its empty
 	    //
	    buildIndexes           (table);
	    System.out.println     ("Empty table");
	    nameIndex.printInOrder ();
	}
    }
}
