package DB.Test;

import NameServer.*;
import DB.*;

// This is a mainline for running the test contained in the Test class
// via RPC.  Depending on how you implement RPC, your version of this
// class may be slightly different from this.


public class TestRPC {

    static final String outputHdr = "\n\n**************************************\n##  DB.Test.TestRPC: ";

    public static void
    main (String argv[])
    {
	String       usageString = "usage: TestRPC <NS-NAME> <LOCAL_NAME> <DBSERVER-NAME> <DB-NAME> <EPOCH> [c|r|e]";
	String       nsName;
	String       localName;
	String       dbsName;
	String       dbName;
	int          epoch;
	char         command;
	NameContextI dbContext, rootContext;
	DatabaseI    db;
	
	//
	// Arg Parsing
	//
	
	try {
	    nsName = argv[0];
	    localName = argv[1];
	    dbsName = argv[2];
	    dbName = argv[3];
	    epoch = Integer.parseInt(argv[4]);
	    command = argv[5].charAt(0);
	} catch (ArrayIndexOutOfBoundsException abe) {
	    System.out.println (usageString);
	    return;
	} catch (NumberFormatException nfe) {
	    System.out.println (usageString);
	    return;
	}
	
	try {
	    // start the server thread and bind the NameServerCP class
	    System.out.println(outputHdr + "starting the local server");
	    RPC.Server.start       (epoch);
	    RPC.Client aClient = new RPC.Client();
	    NameServerCP.bind ("NameServer.NameServer", nsName);	

	    // register the database classes with the dbContext
	    String hashval;
	    System.out.println(outputHdr + "registering classes");

	    hashval = String.valueOf(RPC.ClassID.Cell) + "." + String.valueOf(CellRPC.CLASS_VERSION);
	    RPC.Server.classRegistry.put(hashval, "DB.Cell");
	    System.out.println(outputHdr + "registered DB.Cell");

	    hashval = String.valueOf(RPC.ClassID.Database) + "." + String.valueOf(DatabaseRPC.CLASS_VERSION);
	    RPC.Server.classRegistry.put(hashval, "DB.Database");
	    System.out.println(outputHdr + "registered DB.Database");

	    hashval = String.valueOf(RPC.ClassID.Index) + "." + String.valueOf(IndexRPC.CLASS_VERSION);
	    RPC.Server.classRegistry.put(hashval, "DB.Index");
	    System.out.println(outputHdr + "registered DB.Index");

	    hashval = String.valueOf(RPC.ClassID.Node) + "." + String.valueOf(NodeRPC.CLASS_VERSION);
	    RPC.Server.classRegistry.put(hashval, "DB.Node");
	    System.out.println(outputHdr + "registered DB.Node");

	    hashval = String.valueOf(RPC.ClassID.Table) + "." + String.valueOf(TableRPC.CLASS_VERSION);
	    RPC.Server.classRegistry.put(hashval, "DB.Table");
	    System.out.println(outputHdr + "registered DB.Table");
	
	    // our context object, located in the root context, named DBServer
	    //dbContext = NameServerCP.getRootContext().lookup("DBServer");
	    //if (dbContext==null) {
	    //    System.out.println ("DBServer not found in NameServer");
	    //    return;
	    //}

	    System.out.println(outputHdr + "getting root context");
	    rootContext = NameServerCP.getRootContext();
	    dbContext = (NameContextI)rootContext.lookup(dbsName);
	    RPC.Server.ServerNameContext = dbContext;
	    if(dbContext == null) {
		System.out.println(outputHdr + "DBServer " + dbsName 
				   + " not found in NameServer");
		return;
	    }
	
	    // binding name context CP
	    System.out.println(outputHdr + "binding name context client proxy");
	    NameContextCP.bind(nsName);

	    // create local name context
	    System.out.println(outputHdr + "creating local name context");
	    RPC.Server.LocalNameContext = NameContextCP.rootNameContext.addContext(localName);
	    //dbContext = RPC.Server.ServerNameContext;
	    
	    // lookup the database object, which is stored in our context object
	    // the name of this object is provided as a command line arg
	    System.out.println(outputHdr + "looking up database object");
	    db = (DatabaseI) dbContext.lookup (dbName);
	    if (db==null) {
		System.out.println (outputHdr + "Database object "+dbName+" not found");
		return;
	    }
	    
	    System.out.println(outputHdr + "performing test");
	    // three different types of tests depending on command line arg
	    switch (command) { 
	    
		// create the database
	    case 'c':
		Test.createTable (db, "CS416");
		System.out.println ("Table created; waiting for incoming RPC");
		RPC.Server.join ();
		break;
		
		// read the database
	    case 'r':
		Test.readTable (db, "CS416");
		break;
		
		// empty the database
	    case 'e':
		Test.emptyTable (db, "CS416");
		break;
		
	    default:
		System.out.println (usageString);
		break;
	    }
	} catch(Exception e) {
	    System.out.println(outputHdr + "Error in execution.");
	}
    }
}
