package IPC;

import java.net.DatagramSocket;
import java.io.IOException;


public abstract class Client implements Runnable {

    //
    // Private Data
    //
    private int            maxSeqNum;
    private DatagramSocket comSocket;
    private boolean        up;
    private Thread         listener;
    private CallRec        callRec;

    //
    // Constructor
    //
    public
    Client () 
	throws IOException
    {
	maxSeqNum = 0;
	comSocket = new DatagramSocket ();
	callRec   = new CallRec        ();

	synchronized (this) {
	    up = false;
	    listener = new Thread (this);
	    listener.setDaemon    (true);
	    listener.start        ();
	    while (!up)
		try {
		    wait ();
		} catch (InterruptedException ie) {}
	}
    }

    //
    // Return a new response message
    // IMPLEMENTED BY SUBCLASSES
    //
    protected abstract Message getNewResponseMessage();
    //    {
    //	return new Message();
    //    }

    //
    // Send request and get response which is copied into respMsg
    // SPECIALIZED BY SUBCLASES (for assignement 2)
    //
    public Message
    request (ComHandle dest, Message reqMsg)
	throws IOException,                // unexpected send error (fatal for client)
	       StaleHandleException,       // server com handle is stale
	       ServerUnreachableException  // server unreachable, assume its down
	       
    {
        Message respMsg;
	int     seqNum;
	boolean respReceived;
        int     respType;
	int     respEpoch;
        int     respSeqNum;

	//
	// assemble request-message header
	//
	synchronized (this) {
	    maxSeqNum += 1;
	    seqNum     = maxSeqNum;
	}
	reqMsg.setComHandle   (dest);
	reqMsg.setType        (MessageType.REQUEST);
	reqMsg.setServerEpoch (dest.getEpoch ());
	reqMsg.setSeqNum      (seqNum);
        respMsg = null;

	//
	// send request and wait for response 
	//
	
	respReceived = false;
	for (int i = 0; i <= Params.reqRetryLimit; i++) {
	
	    try {

		//
		// Wait for response (exception thrown if timeout)
		//
		respMsg = callRec.sendReceive (comSocket, dest, seqNum, reqMsg);
	
	        //
	        // Parse response mesage
	        //
	        respType   = respMsg.getType        ();
		respEpoch  = respMsg.getServerEpoch ();
	        respSeqNum = respMsg.getSeqNum      (); 

	        //
	        // check for stale ComHandle error
	        //
	        if (respType == MessageType.STALE_EPOCH) {
	            dest.setEpoch (respMsg.getServerEpoch ());  // update handle
	            throw new StaleHandleException();
	        }

		//
		// check for bad seqnum error
		//
		if (respType == MessageType.BAD_SEQNUM_ERR) {
		    System.out.println ( "Bad seqnum error at client");
		    break;
		}

	        //
	        // Make sure its a valid response
	        //
	        if (respType == MessageType.RESPONSE && 
		    respEpoch == dest.getEpoch() && respSeqNum == seqNum) 
		{
	            respReceived = true;
	            break;
	        }
	        
	        	        
            } catch (IOException ie) {             
                //
                // timeout - start again
                //     
	    } catch (MarshalException me) {	    
	        //
	        // garbled message - start again
	        //
	    }
	}
	
	//
	// if no response from server, throw exeption
	//
	if (respReceived == false) {
	    throw new ServerUnreachableException ();
	}
        
        return respMsg;
    }

    //
    // Listener daemon
    //
    public void 
    run ()
    {
	Message respMsg;
	int     seqNum;

	synchronized (this) {
	    up = true;
	    notify ();
	}

	try {

	    while (up) {

	        respMsg = getNewResponseMessage();

	        respMsg.receive (comSocket);

	        try {
	            seqNum = respMsg.getSeqNum ();
	        } catch (MarshalException me) {
	            System.out.println ("Client listener marshal error");
		    break;
	        }

	        callRec.complete (seqNum, respMsg);
	    }

	} catch (IOException ie) {
	    throw new AssertException ("IOError in Client listener");
	}
    }
}
