package IPC;

import java.lang.Thread;
import java.util.Hashtable;
import java.util.Date;
import java.util.Enumeration;


class RequestEntry {
    int     seqNum;
    int     status;
    Message resp;
    long    expiry;
    
    RequestEntry(int aSeqNum) {
        seqNum  = aSeqNum;
        status  = RequestStatusType.ACTIVE;
        expiry  = 0;
    }
}


class ClientEntry {
    private Hashtable requestList;
    int               lowSeqNum;
    int               highSeqNum;

    ClientEntry ()
    {
        requestList = new Hashtable ();
        lowSeqNum   = 0;
        highSeqNum  = 0;
    }
    
    RequestStatus
    getStatus (int seqNum)
    {
        RequestEntry re;
        
        re = (RequestEntry) requestList.get (new Integer(seqNum));
        if (re != null)
            return new RequestStatus (re.status, re.resp);
        else
            return null;
    }
    
    RequestStatus
    addNew (int seqNum)
    {
        RequestStatus status;
        
        if (seqNum < lowSeqNum)
            status = new RequestStatus (RequestStatusType.STALE, null);
        else {
            if (seqNum <= highSeqNum)
                status = getStatus (seqNum);
            else
                status = null;
            if (status == null) {
                requestList.put (new Integer(seqNum), new RequestEntry(seqNum));
                if (seqNum > highSeqNum)
                    highSeqNum = seqNum;
            }
        }
        
        return status;
    }
    
    void
    setResp (int seqNum, Message resp)
    {
        RequestEntry re;
        
        re = (RequestEntry) requestList.get (new Integer(seqNum));
        Assert.check (re != null);
        re.status = RequestStatusType.SAVED;
        re.resp   = resp;
        re.expiry = (new Date()).getTime() + Params.maxClientRespWait;
    }

    void
    purgeExpired (long asOf)
    {
        RequestEntry re;
        Enumeration  requests;
        int	     newLowSeqNum;
        
        newLowSeqNum = 0;
        requests = requestList.elements ();
        while (requests.hasMoreElements()) {
            re = (RequestEntry) requests.nextElement ();
            if ((re.expiry==0 || re.expiry > asOf) && (newLowSeqNum==0 || re.seqNum<newLowSeqNum))
                newLowSeqNum = re.seqNum;
        }
        
        if (newLowSeqNum > lowSeqNum) {
            lowSeqNum = newLowSeqNum;
            requests = requestList.elements();
            while (requests.hasMoreElements()) {
                re = (RequestEntry) requests.nextElement ();
                if (re.seqNum < lowSeqNum)
                    requestList.remove (new Integer(re.seqNum));
            }
        }
    }
}


public class ClientList implements Runnable {

    private Hashtable clientList;
    private Thread    daemon;

    ClientList ()
    {
	clientList = new Hashtable();
	daemon = new Thread (this);
	daemon.setDaemon    (true);
	daemon.start        ();
    }

    /**
    * Called by server when it receives a new request. 
    * 	@param	client	ComHandle of calling client.
    *	@return 	Request status or null indicating that this is a new request.
    */
    synchronized RequestStatus
    newRequest (ComHandle client, int seqNum) 
    {
	ClientEntry   ce;
        RequestStatus status;

	ce = (ClientEntry) clientList.get (client);
	if (ce == null) {
	    ce = new ClientEntry ();
	    clientList.put (client, ce);
	}
        
        return ce.addNew (seqNum);
    }

    /**
     * Called by ServerRequest when response message completed.
     */
    synchronized void
    saveResponse (ComHandle client, int seqNum, Message respMsg)
    {
	ClientEntry ce;

	ce = (ClientEntry) clientList.get (client);
	Assert.check (ce!=null);
	Assert.check (respMsg!=null);
        ce.setResp (seqNum, respMsg);
    }
    
    /**
     * Daemon garbage collection thread; periodically removes expired entries from 
     * client and request lists.
     */
    public void
    run ()
    {
        Enumeration clients;
        ClientEntry ce;
        long        now;
        
        while (true) {
        
            // wait for next iteration    
            try {
                synchronized (this) {
                    wait (Params.garbageCollFreq);
                }
            } catch (InterruptedException ie) {
		continue;
	    }

        
            // check every entry of client list
            now     = (new Date()).getTime();            
            clients = clientList.elements ();
            while (clients.hasMoreElements()) {
                ce = (ClientEntry) clients.nextElement ();
                ce.purgeExpired (now);
            }
        }
    }
}
