package IPC;

import java.net.InetAddress;
import java.io.DataOutputStream;
import java.io.DataInputStream;
import java.io.IOException;


public class ComHandle {

    private InetAddress address;
    private int         port;
    private int         epoch;

    //
    // Constructors
    //
    
    public
    ComHandle ()
    {
	try {
	    address = InetAddress.getLocalHost();
        } catch (Exception e) {
	    throw new AssertException ("ComHandle threw a " + e.getClass() +
			              "\n    with message: " + e.getMessage());
        }
	this.port  = 0;
	this.epoch = 0;
    }

    public
    ComHandle (int port)
    {
	try {
	    address = InetAddress.getLocalHost();
        } catch (Exception e) {
	    throw new AssertException ("ComHandle threw a " + e.getClass() +
			              "\n    with message: " + e.getMessage());
        }
	this.port  = port;
	this.epoch = 0;
    }

    public
    ComHandle (InetAddress address, int port)
    {
	this.address = address;
	this.port    = port;
	this.epoch   = 0;
    }
    
    public
    ComHandle (InetAddress address, int port, int epoch)
    {
    	this.address = address;
    	this.port    = port;
    	this.epoch   = epoch;
    }
    
    //
    // A very simple (and probably bad) hashCode function
    //
    public int
    hashCode()
    {
	return address.hashCode() + port + epoch;
    }

    //
    // Compare two com handles for equality
    //
    public boolean
    equals(Object thatObject)
    {
	ComHandle that = (ComHandle)thatObject;

        return (this.address.equals(that.address)) &&
	       (this.port  == that.port)    &&
 	       (this.epoch == that.epoch);
    }

    //
    // Return IP address
    //
    public InetAddress
    getAddress () 
    {
	return address;
    }

    //
    // Return port number
    //
    public int
    getPort ()
    {
	return port;
    }
    
    //
    // Return server epoch number
    //
    public int
    getEpoch ()
    {
        return epoch;
    }

    //
    // Set new IP address
    //
    public void
    setAddress (InetAddress address)
    {
	this.address = address;
    }

    //
    // Set new port number
    //
    public void
    setPort (int port)
    {
	this.port = port;
    }
    
    //
    // Set new server epoch number
    //
    public void
    setEpoch (int epoch)
    {
        this.epoch = epoch;
    }

    //
    // Return name as string
    //
    public String
    toString ()
    {
        return "( " + address.toString() + ", " + 
	              Integer.toString(port) + ", " + 
	              Integer.toString(epoch) +
               " )";
    }

    static String
    toStringUnsigned (byte bt)
    {
        if (bt < 0)
	    return Integer.toString (bt+256);
	else
	    return Integer.toString (bt);
    }
    
    //
    // Marshal to data output stream
    //
    public void
    marshal (DataOutputStream dos) 
    {
    	byte addrBytes[];
	Integer addr1, addr2, addr3, addr4;
	int dot1, dot2, dot3;
	String addrName;
    	
	try {
	    //addrBytes = address.getAddress ();
	    //dos.writeInt (addrBytes.length);
	    //dos.write    (addrBytes, 0, addrBytes.length);
	    System.out.print("Marshalling ComHandle: ");
	    addrName = address.getHostAddress();
	    System.out.println(addrName);

	    dot1 = addrName.indexOf(".", 0);
	    dot2 = addrName.indexOf(".", dot1+1);
	    dot3 = addrName.indexOf(".", dot2+1);
	    addr1 = new Integer(addrName.substring(0, dot1));
	    addr2 = new Integer(addrName.substring(dot1+1, dot2));
	    addr3 = new Integer(addrName.substring(dot2+1, dot3));
	    addr4 = new Integer(addrName.substring(dot3+1, addrName.length()));
	    dos.writeInt (addr1.intValue());
	    dos.writeInt (addr2.intValue());
	    dos.writeInt (addr3.intValue());
	    dos.writeInt (addr4.intValue());
	    dos.writeInt (port);
	    dos.writeInt (epoch);
	} catch (IOException ie) {
	    throw new AssertException ("Unexpected IO error in Capability.marshal");
	}
    }
    
    //
    // Unmarshal from data input stream
    //
    public static ComHandle 
    unmarshal (DataInputStream dis)
        throws MarshalException
    {
    	int    addrLen;
        byte   addrBytes[];
	String addrString;
	int addr1, addr2, addr3, addr4;
	int    port;
	int    epoch;
		    
	try {
	    /*addrLen     = dis.readInt ();
	    addrBytes   = new byte [addrLen];
	    dis.readFully (addrBytes, 0, addrLen);
	    addrString  = toStringUnsigned(addrBytes[0])+"."+
	                  toStringUnsigned(addrBytes[1])+"."+
	                  toStringUnsigned(addrBytes[2])+"."+
                          toStringUnsigned(addrBytes[3]); */
			  
	    //dis.readInt();
	    //dis.readInt();
	    addr1 = dis.readInt();
	    addr2 = dis.readInt();
	    addr3 = dis.readInt();
	    addr4 = dis.readInt();
	    addrString = new String(String.valueOf(addr1) + "." +
				    String.valueOf(addr2) + "." +
				    String.valueOf(addr3) + "." +
				    String.valueOf(addr4));
	    port        = dis.readInt ();
	    epoch       = dis.readInt ();
	    return new ComHandle (InetAddress.getByName(addrString), port, epoch);
	} catch (IOException ie) {
	    throw new MarshalException ();
        }	
    }
}
