package IPC;

import java.util.Vector;
import java.util.Enumeration;
import java.util.Date;
import java.io.IOException;
import java.net.DatagramSocket;

//
// This class handles sending IAA messages
//

class Request {

    ComHandle client;
    int       seqNum;
    long      iaaTime;

    Request (ComHandle client, int seqNum)
    {
	this.client  = client;
	this.seqNum  = seqNum;
	this.iaaTime = (new Date()).getTime() + Params.iaaTimeout;
    }
}

public class PendingRequests implements Runnable {

    DatagramSocket comSocket;
    Thread         daemon;
    long           daemonWakeupTime;
    Vector         requests;

    PendingRequests (DatagramSocket comSocket)
    {
	this.comSocket        = comSocket;
	this.daemonWakeupTime = 0;
	this.requests         = new Vector ();

	daemon = new Thread (this);
	daemon.setDaemon    (true);
	daemon.start        ();
    }

    synchronized void
    register (ComHandle client, int seqNum)
    {
	Request request;

	//
	// Save in pending requests queue
	//
	request = new Request (client, seqNum);
	requests.addElement   (request);

	//
	// Start daemon timer unless already running
	//
	if (daemonWakeupTime == 0) {
	    daemonWakeupTime = request.iaaTime;
            synchronized (daemon) {
                daemon.notify ();
            }
	}
    }

    synchronized void
    cancel (ComHandle client, int seqNum)
    {
	Enumeration requestEnum;
	Request     request;
	Request     removeRequest;

	//
	// Scan pending requests, removing cancled and updating daemon sleep time
	//
	removeRequest = null;
	for (requestEnum = requests.elements(); requestEnum.hasMoreElements();) {

	    request          = (Request) requestEnum.nextElement ();
	    removeRequest    = null;
	    daemonWakeupTime = 0;

	    if (request.seqNum==seqNum && request.client.equals(client)) 

		removeRequest = request;

	    else 

		if (daemonWakeupTime==0 || request.iaaTime < daemonWakeupTime)
		    daemonWakeupTime = request.iaaTime;
	}
	if (removeRequest != null)
	    requests.removeElement (removeRequest);
    }

    public void
    run ()
    {
	Enumeration requestEnum;
	Request     request;
	Vector      removeRequests;
	Message     msg;
	long        now;
	long        daemonSleepTime;

	removeRequests = new Vector ();

	for (;;) {

	    //
	    //  Wait until pending request needs IAA message
	    //
	    try {

		while (daemonWakeupTime == 0)
                    synchronized (daemon) {
                        daemon.wait ();
                    }

		synchronized (this) {

		    daemonSleepTime = daemonWakeupTime - (new Date()).getTime();

		    if (daemonSleepTime > 0)
		        wait (daemonSleepTime);

		    if (daemonWakeupTime==0 || daemonWakeupTime>(new Date()).getTime())
		        continue;

	            now              = (new Date()).getTime ();
	            daemonWakeupTime = 0;

	            //
	            // Scan pending requests, sending IAAs and reseting sleep time
	            //
		    removeRequests.removeAllElements ();
	            for (requestEnum = requests.elements(); 
			 requestEnum.hasMoreElements();) 
		    {

		        request = (Request) requestEnum.nextElement ();

		        if (request.iaaTime < now) {

		            //
		            // Send IAA to client
		            //
		            try {
		                msg = new Message ();
		                msg.setComHandle  (request.client);
		                msg.setType       (MessageType.I_AM_ALIVE);
		                msg.setSeqNum     (request.seqNum);
		                msg.send          (comSocket);
		            } catch (IOException ie) {
		  	        throw new AssertException ("IO error sending IAA");
		            }

			    removeRequests.addElement (request);

		        } else 

		            if (daemonWakeupTime==0 || request.iaaTime<daemonWakeupTime)
			        daemonWakeupTime = request.iaaTime;
	            }
		    for (requestEnum = removeRequests.elements(); 
			 requestEnum.hasMoreElements();)
		    {
			request = (Request) requestEnum.nextElement ();
			requests.removeElement (request);
		    }
		    removeRequests.removeAllElements ();
	        }

	    } catch (InterruptedException ie) {
		continue;
	    }
	}
    }
}