package IPC;

import IPC.MarshalException;
import IPC.MessageFullException;

public abstract class ServerRequest implements Runnable {

    protected Server  server;
    protected Message reqMsg;

    protected    
    ServerRequest (Server server, Message reqMsg) 
    {
	this.server = server;
        this.reqMsg = reqMsg;
    }

    //
    // Handle request and build data portion of response
    // IMPLEMENTED BY SUBCLASSES 
    //
    protected abstract Message
    handleRequest ();

    //
    // New request thread starts running here
    //
    public void
    run ()
    {
        Message respMsg;

        try {

	    //
	    // call subclass-specific HandleRequest method	
	    //
	    respMsg = handleRequest ();
	    
	    //
	    // set IPC part of response message header
	    //
	    respMsg.setComHandle  (reqMsg.getComHandle ());
	    respMsg.setType       (MessageType.RESPONSE);
	    respMsg.setSeqNum     (reqMsg.getSeqNum ());

	    //
            // Give message back to server for sending
            //
	    server.sendResponse (respMsg);

	} catch (MarshalException me) {
	    throw new AssertException ("Unexpected marshal error in ServerRequest");
        }
    }
}
