package IPC.Test;

import java.io.IOException;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.net.SocketException;
import java.io.FileInputStream;
import java.io.DataInputStream;
import IPC.ComHandle;
import IPC.Client;
import IPC.Message;
import IPC.MessageFullException;
import IPC.StaleHandleException;
import IPC.ServerUnreachableException;
import IPC.MarshalException;

public class TestClient {

    static Client client;

    public static IPC.ComHandle
    bindFromFile (String fileName)
    {
        ComHandle       serverCH;
        FileInputStream fis;
        DataInputStream dis;

        try {
            fis      = new FileInputStream (fileName);
            dis      = new DataInputStream (fis);
            serverCH = IPC.ComHandle.unmarshal (dis);
            fis.close ();
	    return serverCH;
        } catch (java.io.FileNotFoundException fnf) {
            throw new RPC.NoServerException ();
        } catch (java.io.IOException io) {
            throw new RPC.NoServerException ();
        }
    }

    static void
    testRequest (ComHandle server, int num) 
    {
        Message reqMsg;
  	Message respMsg;
	int     numSqr;

	reqMsg  = new Message ();

	try {

   	    reqMsg.addArgInt           (num);
	    respMsg = client.request   (server, reqMsg);
	    numSqr = respMsg.getArgInt ();
	    System.out.println         (num + " squared is " + numSqr);

	} catch (MessageFullException mfe) {
	    System.out.println ("Unexpected message full");
        } catch (IOException ie) {
	    System.out.println ("Unexpected request IO Exception");
	} catch (StaleHandleException she) {
	    System.out.println ("Unexpected stale handle");
	} catch (ServerUnreachableException sue) {
	    System.out.println ("Server down");
	} catch (MarshalException me) {
	    System.out.println ("Unexpected marshal error for response");
	}
    }
    
    public static void
    main (String argv[])
    {
	String usageString = "usage: TestClient <server-name>";
	ComHandle server;
        
        // initialize
        try {
            client = new Client ();
        } catch (IOException ie) {
	    System.out.println ("Socket error");
	    return;
        }
	try {        
            server = bindFromFile (argv[0]);
	} catch (ArrayIndexOutOfBoundsException abe) {
	    System.out.println (usageString);
	    return;
        }
        
        // perform test
	testRequest (server, 5);
	testRequest (server, 7);
    }
}
