package NameServer;

// This is the server proxy class for the NameServer class.  Any class that
// exports a "class method" via RPC will need one of these, but classes that
// export exclusively "object methods" do not need one.
//
// What you actually put here will depend on how you implement your 
// RPC.ServerProxy class.  So, do that first.  What you see here is what
// is required for my RPC.ServerProxy, you might have something completely
// different.
//
// Note also that binding and registration of the NameServer class will
// be handled differently from all other classes, because all of the other
// classes will use the NameServer for these purposes while the NameServer
// itself will use a file (that's why 'registerClassToFile' does)
//

public class NameServerCP extends RPC.ClientProxy
{
   
 
    static public IPC.ComHandle remoteCH;
 
    static {
	String HashKeyVal;
	//register with the local class registry
	HashKeyVal = getHashKeyVal(RPC.ClassID.NameServer,
				   NameServerRPC.CLASS_VERSION);
	RPC.Server.classRegistry.put(HashKeyVal, "NameServer.NameServer");
	HashKeyVal = getHashKeyVal(RPC.ClassID.NameContext,
				   NameContextRPC.CLASS_VERSION);
	RPC.Server.classRegistry.put(HashKeyVal, "NameServer.NameContext");
	
    }

    static private String getHashKeyVal(int ClassID, int ClassVerNum){
	String VerNum, ClassNum, HashKeyVal;
	ClassNum = String.valueOf(ClassID);
	VerNum = String.valueOf(ClassVerNum);
	return (ClassNum + "." + ClassVerNum);
    }

   static public IPC.ComHandle getClassComHandle(){
       //return new IPC.ComHandle();
       return remoteCH;
    }

    public static void bind(String className, String fileName)
	throws java.io.IOException
    {
	try {
	    remoteCH = bindClassFromFile(className, fileName);
	    //System.out.println("getting root context...");
	    rootNameContext = getRootContext();
	    //System.out.println("got root context.");
	}catch (IPC.ServerUnreachableException sue){
	    System.out.println ("The server " + 
	    			fileName + 
	    			" was not available: "
	    			+ sue.getMessage());
	    return;
	} catch(Exception e) {
	    throw new java.io.IOException();
	} 
    }

    public NameServerCP(RPC.Capability aCap){
	super(aCap);
    }

     public static NameContextI getRootContext() 
	throws IPC.ServerUnreachableException
    {
	NameContextI retObj;
	RPC.Capability respCap;
	NameContextI respNC;

	try{

	   retObj =  (NameContextI)(remoteInvoke(null, RPC.ClassID.NameServer, NameServerRPC.CLASS_VERSION,
			 "getRootContext", NameServerRPC.getRootContext_version, 
			 new String[]{}, new Object[] {}));
	    
	   return retObj;
 
	} catch(Exception e) {
	    throw new IPC.ServerUnreachableException();
	}
    }
}
    
