package RPC;
import java.util.Random;
import IPC.*;
import java.io.DataOutputStream;
import java.io.DataInputStream;
import java.io.IOException;

// Use this class to represent a capability.  I've included the interface
// of the methods I implemented in my solution.  You solution can vary 
// anyway you like.


public class Capability {
    private IPC.ComHandle CH;
    private static int CurrObjectID = 0;
    public int ObjectID;
    public int classID;
    public int classVer;
    
    //
    // Create a new capability
    //	constructor
    //Capability (IPC.ComHandle server)
    //{   
    //    ObjectID = CurrObjectID + 1;
    //    CurrObjectID++;
    //	CH = server;
    //}
    
    Capability (IPC.ComHandle server, int cID, int cVer)
    {   
        ObjectID = CurrObjectID + 1;
	classID = cID;
	classVer = cVer;
        CurrObjectID++;
	CH = server;
    }

    //
    // Copy constructor for capability
    //
    Capability (Capability OtherCap)
    {
	CH = OtherCap.getComHandle();
	ObjectID = OtherCap.getObjId();
    }

    public Capability (IPC.ComHandle inputCH,  int ObjID, int cID, int cVer)
    {
	CH = inputCH;
	classID = cID;
	classVer = cVer;
	ObjectID = ObjID;
    }

    Capability(){
    }

    //
    // This hashCode() method must have two properties:
    //    1. cap1 same as cap2 => cap1.hashCode() == cap2.hashCode()
    //    2. cap1.hashCode() == cap2.hashCode() => cap1 is often not same as cap2
    //
    public int
    hashCode()
    {  //property 1 satisfied by the fact that two identical comhandles hash to the same code
	//and two identical capabilities have the same password pwd
	//property 2 satisfied because using random numbers for the password, and the
	//hash function for the comhandle, it is possible that two different capabilities could
	//hash to the same hashcode

	return CH.hashCode()+classID+classVer+ObjectID;
    }

    //
    // Compare two com handles for equality
    //
    // Returns true iff thatObject is a Capability that is the same as this.
    // Note that thatObject might be a different object than this even if it
    // represents the same capability.
    public boolean
    equals(Object thatObject)
    {
	Capability that = (Capability)thatObject;

	return(this.CH.equals(that.getComHandle()) && (this.ObjectID == that.ObjectID));
    }

    // Extract the server's comHandle from the capability
    public ComHandle
    getComHandle ()
    {
	return CH ; 
    }

    
    public int getObjId(){
	return ObjectID;
    }

    public int getClassId() {
	return classID;
    }

    public int getClassVer() {
	return classVer;
    }

    // write capabilith to dos
    void
    marshal (DataOutputStream dos)
    {
	try{

	// to write a comHandle, call the ComHandle's marshal:
	CH.marshal (dos);
	// to write an int (say i):
	dos.writeInt(ObjectID);
	dos.writeInt(classID);
	dos.writeInt(classVer);
	}catch (IOException ie) {
	    throw new AssertException ("Unexpected IO error in Capability.marshal");
	}
    }
	
    // extract cap from dis
    static Capability
    unmarshal (DataInputStream dis)
	throws MarshalException
    {         
	try{
	IPC.ComHandle comhandle;
	
	int ObjId;
	
	// to read a comHandle:
	      comhandle = ComHandle.unmarshal (dis);
	// to read an int:
	    
	      ObjId = dis.readInt();
	      int cID = dis.readInt();
	      int cVer = dis.readInt();
	      return new Capability(comhandle, ObjId, cID, cVer);
	}catch (IOException ie) {
	    throw new MarshalException();
	}
    }	
}

