 package RPC;

// Use this class to implement any methods or data that are common to all 
// application client proxy classes.  I've included the interface to the methods I
// implemented for my solution as a (hopefully) helpful hint, but do 
// whatever you like.

import IPC.*;
import java.io.*;
import java.net.InetAddress;
import NameServer.*;
import java.lang.reflect.*;
import java.lang.*;

public abstract class ClientProxy extends Proxy {
   
    private static final String outputHdr = "##  RPC.ClientProxy: ";

    private Capability cap;
    public static NameContextI rootNameContext;

    //
    // Public methods
    //

    static{
	rootNameContext = null;
    }

    protected ClientProxy(Capability aCap){
	
	cap = aCap;
	Client.capRegistry.put(cap, this);
	//System.out.println(outputHdr + "Created new CP instance of type: " + this.getClass().getName());
    }

    // Used to bind class to server instance.  Needed only for classes that
    // have "class" (i.e., static) methods.  Returns the ComHandle of server.
    // Instead of className, you might want to pass int ClassID and ClassVer.
    // See comment in ServerProxy.
   public static IPC.ComHandle
   bindClass (String ClassName,  String destServerName)
    throws IOException
    {//ClassID and ClassVer are the arguments to the lookup method on the 
	//nameserver
	//destServerName is the name of the namecontext to lookup.
	//nameSeverName is the name of the nameserver to find the namecontext in

        Client ClientReq = new Client();
	int nameServerClassVerNum;
	//int nameServerPass;
	//long nameServerObjID;
	ComHandle ClassCH;
	Capability nameServerCap;
	RequestMessage reqMsg; 
	ResponseMessage respMsg;
        Object retObj;

	Capability rootNameContextCap;

	try {
	    //System.out.println(outputHdr + "bindClass -> getting Capabilty for a nameContext");
	    Object lookupObj = rootNameContext.lookup(destServerName);
	    //System.out.println(outputHdr + "bindClass: ->lookup returned object of type: " +
	    //		      lookupObj.getClass().getName());
		Capability NCCap2 = ((Proxy)lookupObj).getCap();
		NameContextSP NCSP2 = (NameContextSP)Server.capRegistry.get(NCCap2);
		
		if (NCSP2 == null){
		    //System.out.println(outputHdr + "destination NameContext is on a remote machine.");
		    NameContextCP NCCP2 = (NameContextCP)(Client.capRegistry.get(NCCap2));
		    if(NCCP2 == null){
			System.out.println(outputHdr + "creating new ClientProxy for destination NameContext.");
		    NCCP2 = new NameContextCP(NCCap2);
		    //System.out.println(outputHdr + "Created new NameContextCP object");
		    }else{
			//System.out.println(outputHdr + "bindClass -> Client " + 
			//		   "already has a CP for this nameContext " +
			//		   " in the Client CapRegistry.");
		    }
		     ClassCH = (ComHandle)(NCCP2.lookup(ClassName));
		}else {
		    //System.out.println(outputHdr + "destination NameContext is on the local machine.");
		    NameContext serverNameContext = (NameContext)(NCSP2.getObject());
		     ClassCH = (ComHandle)(serverNameContext.lookup(ClassName));
		}
	

	    return ClassCH;

	} catch (Exception e) {
	    System.out.println(outputHdr + "error binding class to NameServer.");
	    throw new IOException(e.getMessage());
	}
    }

    // Do binding from file.  Should only be needed by NameServer.NameServerCP
    public static IPC.ComHandle
    bindClassFromFile (String className, String fileName)
    throws IOException
    { 
	int nameServerClassVerNum;
	ComHandle nameServerCH = new ComHandle();
	RequestMessage req; 
	ResponseMessage resp;

	try {
	    BufferedReader fin = new BufferedReader(new FileReader(fileName));
	
	    nameServerClassVerNum = Integer.parseInt(fin.readLine());
	    nameServerCH.setAddress(InetAddress.getByName(fin.readLine()));
	    nameServerCH.setPort(Integer.parseInt(fin.readLine()));
	    nameServerCH.setEpoch(Integer.parseInt(fin.readLine()));
	    
	    System.out.println(outputHdr + "read ComHandle "
			       + nameServerCH.toString() + " from file.");
	    
	    return nameServerCH;
	} catch (Exception e) {
	    System.out.println(outputHdr + "error binding class from file.");
	    throw new IOException(); 
	}
    }

    // Remove CP from capRegistry
    public void
    dispose ()
    {
	Server.capRegistry.remove(cap);
    }
    
    //
    // Assign the proxy object's capability; called as part of object construction
    //
    void
    setCap (Capability aCap)
    {
	cap = aCap;
    }

    //
    // Return the client proxy's capability
    //
   public  Capability
    getCap ()
    {
	return cap;
    }
    

protected static Object remoteInvoke(Object target, 
		    int ClassId,  int classVer, 
		    String methodName,  int methodVer,  
		    String[] paramTypes, Object[] args)
{
    try{
	ServerProxy aSP;
	Capability aCap;
	RequestMessage req = new RequestMessage();
	IPC.Message resp;
	String MethodHdr = ".remoteInvoke";
	ResponseMessage ret;
	Object respObj;
	int respClassId, respClassVerNum;
	IPC.ComHandle ClassCH;

	
	//System.out.print(outputHdr + "packing message: \n\tTarget Object Type: ");
	if(target == null) {
	    //System.out.println("null");
	} else {
	    //System.out.println(target.getClass().getName());
	}
	//System.out.print("\n\tClass ID: " + ClassId
	//	 + "\n\tClass Version: " + classVer
	//	 + "\n\tMethod Name: " + methodName
	//	 + "\n\tMethod Version: " + methodVer
	//	 + "\n\tParameter Types(" + paramTypes.length + "): " );
	//for(int i = 0; i < paramTypes.length; i++) {
	//  System.out.print(paramTypes[i]);
	//}
	

	//System.out.println("");

	req.setClassId(ClassId);
	req.setClassVer(classVer);
	//System.out.println(outputHdr + MethodHdr + "Setting classVer...");
	if(target == null) {
	    req.setRequestType(RequestType.CLASS_REQUEST);
	}else{
	    req.setRequestType(RequestType.OBJECT_REQUEST);
	    ClientProxy aCP = (ClientProxy)target;
	    req.setTargetCap(aCP.getCap());
	}
	
	
	req.setMethodVer(methodVer);

	req.addArg(methodName);
	int argCounter = 0;
	//System.out.println("\t" + outputHdr + "Request Arg " + argCounter
	//		   + ": Method Name " + methodName);
	argCounter++;	

	//System.out.println(outputHdr + MethodHdr + ": adding args.length = " + args.length);
	req.addArg(new Integer(args.length));
	//System.out.println("\t" + outputHdr + "Request Arg " + argCounter
	//		   + ": args.length");
	argCounter++;
	//System.out.println(outputHdr + MethodHdr + ": entering for loop, marshalling args...");

	for(int i = 0; i < args.length; i++){
	    req.addArg(paramTypes[i]);
	    //    System.out.println("\t" + outputHdr + "Request Arg " + argCounter
	    //		       + ": paramType " + paramTypes[i]);
	    argCounter++;
	    if(args[i] == null) {
		req.addArg(null);
		//	System.out.println("\t" + outputHdr + "Request Arg " + argCounter
		//		   + ": null");
	    }else if(args[i] instanceof ClientProxy){
		aCap = ((ClientProxy)args[i]).getCap();
		req.addArg(aCap);
		//System.out.println("\t" + outputHdr + "Request Arg " + argCounter
		//		   + ": ClientProxy cap");
		argCounter++;
	    }else if(args[i] instanceof Capability){
		req.addArg(args[i]);
		//System.out.println("\t" + outputHdr + "Request Arg " + argCounter
		//		   + ": Capability");
		argCounter++;
	    }else if (args[i] instanceof ComHandle){
		req.addArg(args[i]); 
		//System.out.println("\t" + outputHdr + "Request Arg " + argCounter
		//		   + ": ComHandle");
		argCounter++;
	    }else if (args[i] instanceof Integer){
		req.addArg(args[i]);
		//System.out.println("\t" + outputHdr + "Request Arg " + argCounter
		//		   + ": Integer " + ((Integer)args[i]).intValue());
		argCounter++;
	    }else if (args[i] instanceof String){
		req.addArg(args[i]);
		//System.out.println("\t" + outputHdr + "Request Arg " + argCounter
		//		   + ": String " + args[i]);
		argCounter++;
	    }else { //object is a server object. Look in objRegistry for sp
		aSP = (ServerProxy)(Server.objRegistry.get(args[i]));
		//System.out.println(outputHdr + "Arg to be added is a real object on this node\n Creating ServerProxy object for this object...");
		if (aSP == null) { //create a new server proxy object for this object
		    String ServerObjClassName = args[i].getClass().getName();
		    Constructor aSPConstructor;
		    
		    String SPName = ServerObjClassName + "SP";
		
		    Class[] objArgsClass = new Class[] {Object.class};
		    Class SPClass = Class.forName(SPName);
		    Object[] objArg = new Object[] {args[i]};

		    aSPConstructor = SPClass.getConstructor(objArgsClass);
		    aSP = (ServerProxy)(aSPConstructor.newInstance(objArg));
		}
		req.addArg(aSP.getCap());	
		//System.out.println(outputHdr + "Request Arg " + argCounter
		//		   + ": ServerProxy cap");
		argCounter++;	
	    }
	}
	//System.out.println(outputHdr + " Testing if target = null");
	//System.out.println(outputHdr + " target = null?" + (target == null));
	if(target == null) {
	    //get class comhandle using java reflection
	    String sClassID = String.valueOf(ClassId);
	    String sClassVer = String.valueOf(classVer);
	    //System.out.println(outputHdr + " in remoteInvoke : ClassRegistry lookup value is..." + sClassID + "." + sClassVer);
	    String lookupValue = new String(sClassID + "." + sClassVer);
	    //System.out.println(outputHdr + " in remoteInvoke : ClassRegistry lookup value is..." + lookupValue);
	    String ClassName = (String)Server.classRegistry.get((String)lookupValue);
	 
	    //System.out.println(outputHdr +" in remoteInvoke: Creating class object for class " + ClassName + "CP");

	    Class aClass = Class.forName(ClassName + "CP");
	    //System.out.println(outputHdr + " in remoteInvoke created class: " + aClass.getName());
	    
	    Method CHMethod = aClass.getMethod("getClassComHandle", null);
	    //System.out.println(outputHdr + "in remoteInvoke found method " + CHMethod.getName());
	    Object classCHObj = CHMethod.invoke(aClass, new Object[]{});
	    //System.out.println(outputHdr + "in remoteInvoke finished invoking method.");
	    ClassCH = (IPC.ComHandle)classCHObj;
	    
	    System.out.println("\n\n\n" + outputHdr + "sending request to " + ClassCH.toString());
	    resp = Client.client.request(ClassCH, req);
	} else {
	    //Target is a capability or a proxy object
	    Capability targetCap; 
	    if(target instanceof ClientProxy){
		targetCap = ((ClientProxy)target).getCap();
	    }else if (target instanceof ServerProxy){
		targetCap = ((ServerProxy)target).getCap();
	    }else{
		//target is already a cap
		targetCap = (Capability)target;
	    }

	    System.out.println(outputHdr + "sending request to: " + (targetCap.getComHandle()).toString() );
	    resp = Client.client.request(targetCap.getComHandle(), req);
	}
	//System.out.println(outputHdr + "remote invoke -> unmarshalling arguments ");
	ret = (ResponseMessage)resp;
	
	
	// check for error messages
	int retType = ret.getResponseType();
	if(retType == ResponseType.BAD_CLASS) {
	    System.out.println(outputHdr + "received bad class error.");
	} else if(retType == ResponseType.BAD_METHOD) {
	    System.out.println(outputHdr + "received bad method error.");
	} else if(retType == ResponseType.BAD_ARG) {
	    System.out.println(outputHdr + "received bad arg error.");
	} else if(retType == ResponseType.BAD_CAP) {
	    System.out.println(outputHdr + "received bad cap error.");
	}
	if(retType != ResponseType.OKAY) {
	    return null;
	} else {
	    //System.out.println(outputHdr + "received response.");
	}


	respObj = ret.getArg();
 
	//unmarshall out args using forCap and responding ClassId;
	if(respObj == null) {
	    System.out.println(outputHdr + "received response of type null\n\n\n");
	    return null;
	}else if (respObj instanceof Capability){
	    //object is a cap for a proxy or an actual object
	    
	    respClassId = ((Capability)respObj).getClassId();
	    respClassVerNum = ((Capability)respObj).getClassVer();
	    //System.out.println(outputHdr + " ClassID, ClassVer = " + 
	    //		       respClassId + ", " + respClassVerNum);
 	    
	    Object realObject;
	    Capability respCap = (Capability)respObj;
	    realObject = ServerProxy.forCap(respCap, respClassId, respClassVerNum);
	    //System.out.println(outputHdr + "remoteInvoke-> forCap created object of type: " + 
	    //	       realObject.getClass().getName());
	    System.out.println(outputHdr + "received response of type " + realObject.getClass().getName());
	    return realObject;
	}else{
	    //response is a comHandle, Integer, String, null
	    System.out.println(outputHdr + "received response of type " + respObj.getClass().getName());
	    return respObj;
	}

	

    }catch(IOException ioe){
	System.out.println(outputHdr + "IOError in remoteInvoke: " + ioe.getMessage());
	return null;
    }catch(StaleHandleException she){
	System.out.println(outputHdr + "StaleHandleException in remoteInvoke: " + she.getMessage());
	return null;
    }catch(ServerUnreachableException sue){
	System.out.println(outputHdr + "ServerUnreachableException in remoteInvoke: " + sue.getMessage());
	return null;
    }catch(IllegalAccessException iae) {
	System.out.println(outputHdr + "IllegalAccessException in remoteInvoke: " + iae.getMessage());
	return null;
    }catch(IllegalArgumentException iarge) {
	System.out.println(outputHdr + "IllegalArgumentException in remoteInvoke: " + iarge.getMessage());
	return null;
    }catch(InvocationTargetException ite) {
	System.out.println(outputHdr + "InvocationTargetException in remoteInvoke: " + ite.getMessage());
	return null;
    }catch(NullPointerException npu) {
	System.out.println(outputHdr + "NullPointerException in remoteInvoke: " + npu.getMessage());
	return null;
    }catch(ExceptionInInitializerError eiie) {
	System.out.println(outputHdr + "ExceptionInInitializerError in remoteInvoke: " + eiie.getMessage());
	return null;
    }catch(Exception e){
	System.out.println(outputHdr + "unexpected error in remoteInvoke: " + e.getMessage() );
	return null;
    }
    
}


}
