package RPC;

import java.io.IOException;
import IPC.AssertException;
import IPC.MessageDatatype;
import IPC.*;
import java.lang.reflect.*;

// This message class extends the IPC.Message class to support passing object
// references in messages.  I have provided some of the implemtation for you, but
// I have left some stuff to you.
//
// There are two methods that you will need to implement yourself: getArg()
// and setArg().  Of course you can call them something different if you
// like, but there are the two methods needed to send and receive objects
// of generic type such as those that are stored in a NameServer.NameContext
//
// Recall that the way you use Message is the following
// On the sending node (e.g., FooCP for IN args):
//	reqMsg.addArg(arg1);
//      reqMsg.addArg(arg2) ...
// Then on receiving side (e.g., FooSP for IN args):
//      arg1 = reqMsg.getArg();
//      arg2 = reqMsg.getArg();



class RPCMessageDatatype {
    static final int CAPABILITY = 1000;
    static final int COMHANDLE  = 1001;
    static final int NULL       = 1002;
    static final int TYPED      = 1003;
    static final int BOOLEAN    = 1004;
    //in IPC.MessageDatatype Int  = 1
    //String = 2
    //capability = 3
    
}


//
// Common to both RequestMessage and ResponseMessage
//

public class Message extends IPC.Message {

private static final String outputHdr = "##  RPC.Message: ";



    // [addArg]
    //     Marshal an arg of any type.  Encode object type and capability
    //     in the message.  You can encode type, for example, by storing
    //     its name as a string in the message (or do something different
    //     if you like).
    //
    public void
    addArg (Object arg)
    throws IPC.MessageFullException
    {
	String classname;
	try{
	    //System.out.println(outputHdr + "resetting baos...");
	    baos.reset ();
	    //System.out.println(outputHdr + "testing if arg null...");
	    if(arg == null){
		//System.out.println(outputHdr + "arg added is null");
		addArgComHandle(null);
		return;
	    }else{
		classname = arg.getClass().getName();
		//System.out.println(outputHdr + " addArg -> classname = " + classname);
		
		if (classname.equals("RPC.Capability")) {
		   
		    addArgCap((Capability)arg);
		}else if(classname.equals("IPC.ComHandle")){
		 
		    addArgComHandle((IPC.ComHandle)arg);
		}else if (classname.equals("java.lang.String")){
		   
		    addArgInt(RPCMessageDatatype.TYPED);
		    addArgString((String)arg);
		}else if (classname.equals("java.lang.Integer")) {
		   
		    addArgInt(RPCMessageDatatype.TYPED);
		    Integer myInt = (Integer)arg;
		    addArgInt(myInt.intValue());
		}else if (classname.equals("java.lang.Boolean")) {
		    addArgInt(RPCMessageDatatype.BOOLEAN);
		    Boolean myBool = (Boolean)arg;
		    int boolVal;
		    if(myBool.booleanValue()) {
			boolVal = 1;
		    } else {
			boolVal = 0;
		    }
		    addArgInt(boolVal);
		}else if (arg instanceof Proxy) { 
		    //argument is a proxy object
		  
		    Capability aCap;
		    if (arg instanceof ClientProxy){
			ClientProxy aCP = (ClientProxy)arg;
			aCap =  aCP.getCap();
		    }else if(arg instanceof ServerProxy){
			//object is a serverProxy object on the current machine
			ServerProxy aSP = (ServerProxy)arg;
			aCap = ServerProxy.getCap(aSP);
		    }else{
			System.out.println(outputHdr + "addArg error: unexpected object type."); 
			throw new MarshalException();
		    }
		    addArgCap(aCap);
		}else{ 
		    // arg is on the local machine, so try to create a 
		    // server proxy for it.
		    //System.out.println(outputHdr + "arg is object on local machine");
		    //System.out.println(outputHdr + " creating classobject for " +
		    //		       classname + "SP");
		    //String argType = arg.getClass().getName();
		    Class ArgSP = Class.forName(classname + "SP");
		    //System.out.println(outputHdr + " creating constructor for this class...");
		    Constructor ArgSPCons = ArgSP.getConstructor(new Class[] {Object.class});
		    //System.out.println(outputHdr + " creating SP object for this class...");
		    Object newArgSP = ArgSPCons.newInstance(new Object[] {arg});
		   
		    //System.out.println(outputHdr + " creating capability for this class...");
		    Capability newArgCap = ((ServerProxy)newArgSP).cap;
		    addArgCap(newArgCap);
		}
	    }
	//System.out.println(outputHdr + "Datacursor is " + dataCursor);
	} catch (Exception e){
	    addArgInt(RPCMessageDatatype.NULL);
	    System.out.println(outputHdr + "addArg Unexpected error: " +  e.getMessage());
	}  
    }

    //
    // [getArg]
    //     Unmarshal a typed object returning the object or, for base types, a wrapper
    //     object (e.g., Integer for int).
    //
    public Object
    getArg ()
    throws MarshalException
    {
	
	int tag;
	try{
	    tag = getArgInt ();
	    //tag = RPCMessageDatatype.TYPED;
	    //System.out.println(outputHdr + "Arg type is " + tag);
	 if(tag == RPCMessageDatatype.CAPABILITY){ 
	     //System.out.println(outputHdr + "unmarshalling Capability");
	     //go to ServerProxy CapRegistry to get
	     //ServerProxy object. If no ServerProxy Object for this cap,
	     //then cap is for a clientProxy. Check CPCapReg, if it doesn't exist
	     //then create a new ClientProxy object for the capability

	     Capability aCap = getArgCap();
	     //return of either a real Object, or a ClientProxy object
	     //is handled by the ServerProxy or the ClientProxy
	     return aCap; 
	 }else if (tag == RPCMessageDatatype.COMHANDLE){
	     //System.out.println(outputHdr + "unmarshalling ComHandle");
	     ComHandle aCH = getArgComHandle();
	     return aCH;
	 }else if (tag == RPCMessageDatatype.BOOLEAN){
	     int typeTag = dis.readInt();
	     int boolVal = dis.readInt();
	     if(boolVal == 1) {
		 return new Boolean("true");
	     } else {
		 return new Boolean("false");
	     }
	  }else if (tag == RPCMessageDatatype.TYPED) {
	     
	     int  typeTag = dis.readInt();
	     //System.out.println(outputHdr + "Type tag is " + typeTag);
	      if (typeTag == IPC.MessageDatatype.INT){
		  int argint = dis.readInt();
		  //System.out.println(outputHdr + " arg value is: " + argint);
		  return new Integer(argint);
	      }else if (typeTag == IPC.MessageDatatype.STRING){
		  //System.out.println(outputHdr + "in getArg, TypeTag is STRING");
		  int len;
		  byte data[];
		  //System.out.println(outputHdr + "unmarshalling type String");
		  len = dis.readInt();
		  data = new byte[len];
		  //System.out.println(outputHdr + "reading string of length " + len);
		  dis.readFully(data, 0,len);
		  String myString = new String(data, 0, len);
		  //System.out.println(outputHdr + myString);
		  return myString;
	      }else {
		  throw new MarshalException();
	      }
		  
	      //argument is an integer or string
	      
	     }else if (tag == RPCMessageDatatype.NULL) {
		 //System.out.println(outputHdr + "getArg: tag in Message was null.");
		 return null;
	     }else {
		 System.out.println(outputHdr + "Marshal exception in getArg");
		 throw new MarshalException();
	     }
	} catch (IOException ie) {
	    System.out.println(outputHdr + "IOException in getArg. Throwing new MarshalException...: " + ie.getMessage());
	    throw new MarshalException();
	}
	// To read from the message do the following:
	// 2. tag will tell you the type of message (e.g., NULL or TYPED)
	// 3. call getArgXXX() as many times as you like
	// 4. Recall that the object that you return will either be a client proxy
	//    for object named by the next CAP in the message, if the named object
	//    is stored on another node, or the real object, if the object is stored
	//    on this node.  Also note that if you want this method to support passing
	//    things that aren't objects such as int's, you can but it requires wrapping
	//    them in an object (an Integer in the case of ints).  If you don't want
	//    to support this its fine.  If you do the bonus, you'll probably want
	//    to do this.  If you don't, you probably won't.
    }

    //
    // Helper methods to marshal/unmarshal various types of objects
    //

    public void 
    addArgComHandle (IPC.ComHandle comHandle)
    	throws IPC.MessageFullException
    {
	try {
    	    baos.reset ();
	    if (comHandle!=null) {
    	        addArgInt      (RPCMessageDatatype.COMHANDLE);
    	        comHandle.marshal (dos);
		addFromDOS();
	    } else {
	        addArgInt      (RPCMessageDatatype.NULL); 
	    }
	} catch(Exception e) {
	    throw new AssertException ("Exception in addArgComHandle");
	}
    }
	
    public IPC.ComHandle
    getArgComHandle ()
    	throws IPC.MarshalException
    {
	try {
	    dis.readInt(); // need this.  don't know why.
	    int type = dis.readInt();
	    //int type = getArgInt();
	    switch (type) {
	    case RPCMessageDatatype.COMHANDLE:
		return IPC.ComHandle.unmarshal (dis);
	    case RPCMessageDatatype.NULL:
		return null;
	    default:
		throw new IPC.MarshalException ();
	    }
	} catch (Exception ie) {
	    throw new IPC.MarshalException ();
	}
    }
    
    protected Capability
    getCapAt (int offset)
        throws IPC.MarshalException
    {
	if ( (offset+4) > getPacket().getLength() )
	    throw new IPC.MarshalException ();
	bais_hdr.reset ();
	try {
	    dis_hdr.skipBytes (offset);
	    return Capability.unmarshal (dis_hdr);
	} catch (IOException ie) {
	    throw new IPC.MarshalException ();
	}
    }
     
   // Capability to the object being accessed
    protected void
    setCapAt (int offset, Capability cap)
    {
	byte[] bytes;
	byte[] buffer;
	
	baos.reset  ();
	cap.marshal (dos);
	bytes  = baos.toByteArray ();
	buffer = getBuffer();
	for (int i=0; i<baos.size(); i++)
	    buffer[offset+i] = bytes[i];
    }
    
    public Capability
    getArgCap ()
        throws IPC.MarshalException
    {
	int requestType;
	
	try {
	    //dis.readInt();
	    requestType = getArgInt ();
	    if (requestType == RPCMessageDatatype.NULL)
		return null;
	    else if (requestType == RPCMessageDatatype.CAPABILITY)
		return Capability.unmarshal (dis);
	    else
	        throw new IPC.MarshalException ();
	} catch (Exception ie) {
	    throw new IPC.MarshalException ();
	}
    }

    //Capability passed as an object being passed into the  function.
    public void
    addArgCap (Capability cap)
        throws IPC.MessageFullException
    {
	try {
	    baos.reset ();
	    if (cap==null) {
		addArgInt (RPCMessageDatatype.NULL);
	    } else {
	    	addArgInt (RPCMessageDatatype.CAPABILITY);
		cap.marshal (dos);
		addFromDOS();
	    }
	} catch (Exception e) {
	    throw new AssertException ("Exception in addArgCap");
	}
    }
}
