package RPC;

import NameServer.*;
import java.util.Hashtable;
import java.io.IOException;

// This class is a fully implemented version taken from my solution.  If
// your solution works like mine, you should not need to change this at all.
// On the other hand, you are free to change it in any way you like.


public class Server extends IPC.Server implements Runnable {

    static public Server    server;
    static Thread    listener;
    static public Hashtable objRegistry;
    static public Hashtable capRegistry;
    static public Hashtable classRegistry; 

    //the namecontext for this node on the server, under which all
    //classes and objects for this server will be registered
    static public NameContextI ServerNameContext;
    static public NameContextI LocalNameContext;

    //classRegistry -> mapping of classID/Ver to ClassName

    // static constructor is called when this class is loaded
    static {
        objRegistry   = new Hashtable ();
        capRegistry   = new Hashtable ();
        classRegistry = new Hashtable ();
    } 

    Server (int epoch)
        throws IOException
    {
    	super (epoch);
	server = this;
    }

    //
    // Public methods
    //

    // Creates a listener, starts the thread running in the 'start()' method
    // of the server object, which is implemented in IPC.Server.
    public static void 
    start (int epoch) 
    {
	try {
            new Client ();
            new Server (epoch);
            listener = new Thread (server);
	    listener.setDaemon (true);
            listener.start     ();
	} catch (IOException ex) {
	    throw new IPC.AssertException ("IOException in RPC.Server.start()");
	}
    }

    // The calling thread blocks until the listener thread terminates.  Use
    // this to have your mainline program thread block.
    public static void 
    join ()
    {
	try {
	    listener.join ();
	} catch (java.lang.InterruptedException e) {
	}
    }

    // Alternate call to start listener.
    public void
    run ()
    {
	start ();
    }
    
    //
    // Protected methods
    //

    // Called by IPC.Server.start() to get a new ServerRequest.
    protected IPC.ServerRequest
    getNewServerRequest (IPC.Message reqMsg)
    {
	return new ServerRequest (this, reqMsg);
    }
    
    // Called by IPC.Server.start() to get a new request Message.
    protected IPC.Message
    getNewRequestMessage ()
    {
    	return new RequestMessage ();
    }
}
