package RPC;

// Use this class to implement any routines that will be common to all
// server proxies for application classes.  I have included some methods
// here as suggestions of what you might want to include.  Change anything
// you like.
import RPC.Server;
import java.io.*;
import java.net.*;
import java.lang.reflect.*;

public abstract class ServerProxy extends Proxy {

    static final String outputHdr = "##  RPC.ServerProxy: ";

    public Capability cap;
    Object obj;
    static public NameServer.NameContextI currentContext;

    // constructor
    protected ServerProxy(Object o)
    {
	//System.out.println(outputHdr + "creating new ServerProxy object...");
	obj = o;
	//System.out.println(outputHdr + "getting comHandle and creating new cap");

	// use java reflection to get classID and classVersion
	String className;
	Class ThisClassRPC;
	Field classVerField;
	int thisClassVerNum;
	int lastDot;
	String classNameShort;
	Field classIDField;
	int classID;
	try {
	    className = o.getClass().getName();
	    ThisClassRPC = Class.forName(className + "RPC");
	    classVerField = ThisClassRPC.getField("CLASS_VERSION");
	    thisClassVerNum = classVerField.getInt(ThisClassRPC);
	    
	    lastDot = className.lastIndexOf(".");
	    classNameShort = className.substring(lastDot+1, className.length());
	    classIDField = ClassID.class.getDeclaredField(classNameShort);
	    classID = classIDField.getInt(ClassID.class);
	} catch(Exception e) {
	    classID = -1;
	    thisClassVerNum = -1;
	    System.out.println(outputHdr + "error in ServerProxy constructor: " + e.getMessage());
	}
	
	cap = new Capability(Server.server.getComHandle(), classID, thisClassVerNum);
	//System.out.println(outputHdr + "putting cap in Server cap registry");
	RPC.Server.capRegistry.put(cap, this);
	//System.out.println(outputHdr + "putting object in object registry...");
	RPC.Server.objRegistry.put(o, this);
	

	//System.out.println(outputHdr + "adding object to the current Name Context...");
	Server.LocalNameContext.add(String.valueOf(cap.getObjId()), cap);
	//System.out.println(outputHdr + "returning to calling method...");
    }

    // return the object
    public Object getObject()
    {
	return obj;
    }


    public Capability getCap(){
	return cap;
    }

    //
    // Find or create capability for target object
    //
    static Capability
    getCap (Object obj)
    {
	Capability c = null;

	// Return the capability for the given application object using
	// tables stored in the RPC.Server object.
	c = ((ServerProxy)RPC.Server.objRegistry.get(obj)).cap;

	// If there's no capability for the object, create a server proxy
	// and capability for the object.
	if(c == null) {
	    try {
		Class ObjSP = Class.forName(obj.getClass().getName() + "SP");
		Constructor ObjSPCons = ObjSP.getConstructor(new Class[] {Object.class});
		Object newObj = ObjSPCons.newInstance(new Object[] {obj});
		Field objCap = ObjSP.getField("cap");
	    c = (Capability)(objCap.get(newObj));
	    } catch(Exception e) {
		c = null;
	    }
	}
	return c;
    }

    //
    // Find target object from capability
    //
    static Object
    forCap (Capability cap, int classID, int classVer)
    {
	if(cap == null) {
	    return null;
	}

	// check the server cap registry
	ServerProxy sp;
	sp =  (ServerProxy)(Server.capRegistry.get(cap));
	if( sp != null) {
	    return sp.getObject();
	}

	// check the client cap registry
	ClientProxy cp;
	cp = (ClientProxy)(Client.capRegistry.get(cap));
	if( cp != null ) {
	    return cp;
	}

	// create a new CP for the cap
	try {
	    String className;
	    className = (String)(Server.classRegistry.get(String.valueOf(classID) 
							  + "." 
							  + String.valueOf(classVer)));
	    //System.out.println(outputHdr + " Creating class object for " + className + "CP");
	    Class ObjCP = Class.forName(className + "CP");
	    Constructor ObjCPCons = ObjCP.getConstructor( new Class[] {Capability.class} );
	    
	    Object retObj =  ObjCPCons.newInstance(new Object[] {cap});
	    //System.out.println(outputHdr + "Object Created: " + retObj.getClass().getName());
	    return retObj;
	} catch(Exception e) {
	    System.out.println(outputHdr + " in forCap id " + classID + " ver " + classVer + ": " + e.getMessage());  
	    return null;
	}
    }
    
    //
    // Remove object from server cap and object (and class) registries
    //
    public static void
    dispose (Object obj)
    {
	ServerProxy sp;
	sp = (ServerProxy)(RPC.Server.objRegistry.get(obj));
	if(sp != null) {
	    RPC.Server.objRegistry.remove(obj);
	    RPC.Server.capRegistry.remove(sp.cap);
	}
    }
    
    //
    // Register class for class-based RPC.  Not needed if all remote invocations
    // to class are object-based RPC.  The way I've done this is that
    // the caller passes in the className of the application class to
    // be registered.  This assumes that you can somehow get the ClassID
    // and ClassVersionNum etc. from the className (e.g., you could have
    // a registry for classes).  This approach is helpful if you do the
    // bonus and use Java reflection.  Feel free, however, to pass in
    // the ClassID and ClassVersionNum or any other information you feel
    // is needed.
    //
    public static void
    registerClass(String className)
	throws IPC.ServerUnreachableException
    {
	// adds className and server comHandle  to the name context 
	// ncName in the name server described by fileName
	
	System.out.println(outputHdr + "registering Class " 
			   + className + " on NameContext ");
			   //+ ncName);
	
	try {
	    //NameServer.NameContextI rootNC = ClientProxy.rootNameContext;
	    NameServer.NameContextI currNC = Server.ServerNameContext;
	    Object classObj;

	    classObj = currNC.lookup(className);
	    if(classObj == null) {
		//System.out.println(outputHdr + "adding " + className 
		//		   + " to NameContext "); // + ncName);
		currNC.add(className, Server.server.getComHandle());
	    }
	    //System.out.println(outputHdr + "finished registering.");
	} catch(Exception e) {
	    System.out.println(outputHdr + "error in registering.");
	    throw new IPC.ServerUnreachableException();
	}
    }

    // Register class using a file instead of the NameServer.  Typically
    // used only by the NameServerSP.  
    public static void
    registerClassToFile (String className,  String fileName)
    {
	IPC.ComHandle ch = Server.server.getComHandle();
	try {
	    //System.out.println(outputHdr + "registering Class "
	    //		       + className + " to file "
	    //		       + fileName + "...");

	    PrintWriter fout = new PrintWriter( new FileOutputStream(new File(fileName)) );

	    //System.out.println(outputHdr + "writing: \n\tClass Version: "
	    //		       + NameServer.NameServerRPC.CLASS_VERSION 
	    //		       + "\n\tComHandle: " + ch.toString());

	    fout.write(NameServer.NameServerRPC.CLASS_VERSION + "\n"
		       + ch.getAddress().getHostName() + "\n" 
		       + ch.getPort() + "\n"
		       + ch.getEpoch() + "\n");
	    fout.close();
	    //System.out.println(outputHdr + "finished registering.");
	} catch(Exception e) {
	    System.out.println(outputHdr + "error in registering.");
	}
    }

    //
    // Handle server side of a remote invocation (called by ServerRequest)
    //
    // public static void 
    public static void
    handleRequest (RequestMessage req, ResponseMessage resp, String className)
    {      
	String methodName;
	String paramType;
	Integer numArgsI;
	int numArgs;
	Method reqMethod;
	Object arg;
	Object reqSP;
	Object reqCP;
	Class ReqClass;
	Class ThisClassRPC;

	try {
	    System.out.println(outputHdr + "handling request from Class "
			       + className);
	   
	    // check the class version number
	    int classVerNum = req.getClassVer();
	    ThisClassRPC = Class.forName(className + "RPC");
	    Field classVerField = ThisClassRPC.getField("CLASS_VERSION");
	    int thisClassVerNum = classVerField.getInt(ThisClassRPC);
	    if(classVerNum != thisClassVerNum) {
		resp.setResponseType(ResponseType.BAD_CLASS);
		System.out.println("\tBad class version number " 
				   + classVerNum);
		return;
	    }
	    //System.out.println("\tClass Version: " + classVerNum);
	    
	    methodName = (String)(req.getArg());
	    System.out.println("\tMethod: " + methodName);
	    numArgsI = (Integer)(req.getArg());
	    numArgs = numArgsI.intValue();
	    // numArgs = 0;
	    System.out.println("\tNumber of args: " + numArgs);
	    
	    Class[] reqTypes = new Class[numArgs];
	    Object[] reqArgs = new Object[numArgs];
	    //System.out.println("\tArguments: \n");
	    for(int i = 0; i < numArgs; i++) {
		paramType = (String)(req.getArg());
		//System.out.println(outputHdr + " paramType = " + paramType);
		arg = req.getArg();
		//System.out.println(outputHdr + " Handling" + paramType + "now");
		if(paramType.equals("java.lang.Integer")) {
		    reqTypes[i] = Integer.class;
		    reqArgs[i] = arg;
		    //System.out.print("java.lang.Integer ");
		} else if(paramType.equals("java.lang.String")) {
		    //System.out.println(outputHdr + "putting String.class into reqTypes");
		    reqTypes[i] = String.class;
		    //System.out.println(outputHdr + "putting arg into reqTypes[" + i + "]");
		    reqArgs[i] = arg;
		    //System.out.print("java.lang.String ");
		} else if(paramType.equals("IPC.ComHandle")) {
		    reqTypes[i] = IPC.ComHandle.class;
		    reqArgs[i] = arg;
		    //System.out.println("ComHandle ");
		} else if(paramType.equals("java.lang.Object")) {
		    reqTypes[i] = Object.class;
		    reqArgs[i] = arg;
		    //System.out.println("Object ");
		} else if(paramType.equals("java.lang.Boolean")) {
		    reqTypes[i] = Boolean.class;
		    reqArgs[i] = arg;
		    //System.out.println("Boolean");
		} else { // arg instanceof Capability
		    //System.out.println(outputHdr + " paramType else: " + arg.getClass().getName() + ", " + paramType);
		    reqSP = Server.capRegistry.get((Capability)arg);
		    reqCP = Client.capRegistry.get((Capability)arg);
		    //ReqClass = Class.forName(paramType + "I");
		    if(reqSP != null) {
			Class ReqSPClass = reqSP.getClass();
			Method reqSPMethod = ReqSPClass.getMethod("getObject", 
								  new Class[] {});
			Object reqSPObj = reqSPMethod.invoke(reqSP, new Object[] {});
			reqArgs[i] = reqSPObj;
			reqTypes[i] = reqSPObj.getClass();
			//System.out.println(reqSPObj.getClass().getName() + " ");
		    } else if(reqCP != null) {
			reqArgs[i] = reqCP;
			//reqTypes[i] = reqCP.getClass();
			String reqCPObjName = reqCP.getClass().getName();
			int CPpos = reqCPObjName.lastIndexOf("CP");
			String reqObjName = reqCPObjName.substring(0,CPpos);
			Class reqObjClass = Class.forName(reqObjName + "I");
			reqTypes[i] = reqObjClass;
			//System.out.println(reqObjClass.getName() + " ");
		    } else {
			Class ReqClassCP = Class.forName(paramType + "CP");
			Class[] objArgsClass = new Class[] {Capability.class};
			Object[] objArgs = new Object[] {arg};
			Constructor ReqClassCPCons = ReqClassCP.getConstructor(objArgsClass);
		        Object newCP = ReqClassCPCons.newInstance(objArgs);
			reqArgs[i] = newCP;
			//reqTypes[i] = newCP.getClass();
			String newCPObjName = newCP.getClass().getName();
			int CPpos = newCPObjName.lastIndexOf("CP");
			String newObjName = newCPObjName.substring(0,CPpos);
			Class newObjClass = Class.forName(newObjName + "I");
			reqTypes[i] = newObjClass;
			//System.out.print(newObjClass.getName() + " ");
		    }
		}
	    }
	    //System.out.println("<endofargs>");
	    
	    Capability targetCap;
	    int methodReqType = req.getRequestType();
	    if(methodReqType == RequestType.CLASS_REQUEST) {
		ReqClass = Class.forName(className);
		targetCap = null;
	    } else {
		ReqClass = Class.forName(className + "I");
		//System.out.println(outputHdr + " HandleRequest: getting target cap...");
		targetCap = req.getTargetCap();
		//System.out.println(outputHdr + " got TargetCap: " + (targetCap.getComHandle()).toString());
		//System.out.println(targetCap.getClassId());
	    }

	    //System.out.println(methodName);
	    //for(int n=0; n<reqTypes.length; n++) {
	    //	System.out.println(reqTypes[n]);
	    //}
	    reqMethod = ReqClass.getMethod(methodName, reqTypes);
	    //System.out.println("got method");
	    // check the method version
	    int methodVerNum = req.getMethodVer();
	    Field methodField = ThisClassRPC.getDeclaredField(reqMethod.getName()
							 + "_version");
	    int reqMethodVerNum = methodField.getInt(ThisClassRPC);
	    int thisMethodVerNum = methodField.getInt(ThisClassRPC);
	    if(reqMethodVerNum != thisMethodVerNum) {
		resp.setResponseType(ResponseType.BAD_METHOD);
		System.out.println("\tBad method version number " + reqMethodVerNum);
		return;
	    }

	    Object respObj;
	    if(targetCap == null) { // class method
		System.out.println("\tType: Class Request");
		respObj = reqMethod.invoke(ReqClass, reqArgs);
		if(respObj == null) {
		    System.out.println("\tResponse Type: null");
		    resp.addArg(null);
		} else {
		    
		    String respName = respObj.getClass().getName();
		    System.out.println("\tResponse Type: " + respName);
		    resp.addArg(respObj);
		    // set the classID and classVerNum
		    try {
			Class RespRPC = Class.forName(respName + "RPC");
			Field respVerField = RespRPC.getDeclaredField("CLASS_VERSION");
			int respVer = respVerField.getInt(RespRPC);
			//System.out.println(outputHdr + "Class Version " + respVer);
			
			int lastDot = respName.lastIndexOf(".");
			String respNameShort = respName.substring(lastDot+1, respName.length());
			Field respIDField = ClassID.class.getDeclaredField(respNameShort);
			int respID = respIDField.getInt(ClassID.class);
			//System.out.println(outputHdr + "Class ID " + respID);
			
			resp.setClassId(respID);
			resp.setClassVer(respVer);
		    } catch (Exception e) {
			// respObj is not one of the application classes
			// but we really don't care here.
			resp.setClassId(-1);
			resp.setClassVer(-1);
		    }
		}
	    }
	    else { // object method
		Object targetObj;
		System.out.println("\tType: Object Request");

		ServerProxy targetSP = (ServerProxy)(Server.capRegistry.get(targetCap));
		//System.out.println(outputHdr + " targetSP is of type: " + targetSP.getClass().getName()); 
		targetObj = targetSP.getObject();
		//System.out.println(outputHdr + " targetObject is of type: " + targetObj.getClass().getName());

		respObj = reqMethod.invoke(targetObj, reqArgs);
		if(respObj == null) {
		    System.out.println("\tResponse Type: null");
		    resp.addArg(null);
		} else {
		    String respName = respObj.getClass().getName();
		    System.out.println("\tResponse Type: " + respName);
		    resp.addArg(respObj);
		    try {
			Class RespRPC = Class.forName(respName + "RPC");
			Field respVerField = RespRPC.getDeclaredField("CLASS_VERSION");
			int respVer = respVerField.getInt(RespRPC);
			//System.out.println(outputHdr + "Class Version " + respVer);
			
			int lastDot = respName.lastIndexOf(".");
			String respNameShort = respName.substring(lastDot+1, respName.length());
			Field respIDField = ClassID.class.getDeclaredField(respNameShort);
			int respID = respIDField.getInt(ClassID.class);
			//System.out.println(outputHdr + "Class ID " + respID);
			
			resp.setClassId(respID);
			resp.setClassVer(respVer);
		    } catch (Exception e) {
			// respObj is not one of the application classes
			resp.setClassId(-1);
			resp.setClassVer(-1);
		    }
		}
	    }

	    resp.setResponseType(ResponseType.OKAY);

	} catch(NoSuchMethodException ncme) {
	    resp.setResponseType(ResponseType.BAD_METHOD);
	} catch(Exception e) {
	    resp.setResponseType(ResponseType.BAD_ARG);
	    System.out.println(outputHdr + "Unexpected error in handle request: " + e.getMessage());
	}
    }
}
